import sys,os
import re

from pyral import Rally, RallyRESTAPIError
from dbeif.connection import Connection
from dbeif.utils.eif_exception import ConfigurationError, OperationalError
from dbeif.utils.etracer import Tracer

__version__ = '1.1.0'

class AgileCentralConnection(Connection):
    def __init__(self, config, logger, metrics_recorder):
        self.config = config
        self.mr     = metrics_recorder
        super().__init__(logger)

        self.internalizeConfig(config)
        self.log.info("Rally WSAPI Version %s" % self.rallyWSAPIVersion())
        self.integration_other_version = ""
        self.username_required = True
        self.password_required = True
        self.field_categories = {}
        self.pi_states_map = {}
        self._project_cache = {}

    def name(self):
        return 'dbeif.AgileCentralConnection'

    def version(self):
        return __version__

    def rallyWSAPIVersion(self):
        return "v2.0"

    def set_integration_header(self, header_info):
        self.integration_name    = header_info['name']
        self.integration_vendor  = header_info['vendor']
        self.integration_version = header_info['version']
        if 'other_version' in header_info:
            self.integration_other_version = header_info['other_version']

    def setSourceIdentification(self, other_name, other_version):
        self.other_name = other_name
        self.integration_other_version  = other_version

    def get_custom_headers(self):
        custom_headers =  {}
        custom_headers['name']    = "AgileCentral Database Connector for %s" % self.other_name
        custom_headers['vendor']  = "CA Technologies"
        custom_headers['version'] = self.version()
        if self.integration_other_version:
            spoke_versions = "%s - %s " % (self.version(), self.integration_other_version)
            custom_headers['version'] = spoke_versions
        return custom_headers


    def internalizeConfig(self, config):
        super().internalizeConfig(config)

        server = config.get('Server', 'rally1.rallydev.com')
        if 'http' in server.lower() or '/slm' in server.lower():
            self.log.error(self, "AgileCentral URL should be in the form 'rally1.rallydev.com'")
        self.server = server

        self.url = "https://%s/slm" % server
        self.port = config.get('Port', '443')
        self.apikey          = config.get("APIKey", config.get("API_Key", None))
        self.workspaces      = config.get("Workspaces", None)
        self.workspace_name  = self.workspaces[0]['Workspace']
        self.project_name    = None
        self.entities = [item.strip() for item in config.get('Entities', "").split(',')]
        self.fetch           = config.get("Fetch", None)
        self.base_query      = config.get('BaseQuery', '')
        self.selectors       = config.get('Selectors', [])
        self.restapi_debug   = config.get("Debug", False)
        self.restapi_logger  = self.log
        if self.fetch.count("\n"):
            self.fetch = "".join([line.strip() for line in self.fetch.split("\n")])

        self.proxy = None
        if self.proxy_server:
            self.proxy  = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
            if self.proxy_username and self.proxy_password:
                self.proxy  = "%s://%s:%s@%s:%s" % (self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)

        valid_config_items = ['Server', 'Port', 'APIKey','API_Key','Workspaces', 'Username','Password',
                              'ProxyProtocol', 'ProxyServer','ProxyPort','ProxyUser','ProxyUsername', 'ProxyPassword',
                              'BaseQuery','Selectors','Fetch','Entities', 'ResolveUser', 'Selectors'
                              'Debug']

        invalid_config_items = [item for item in config.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "AgileCentral section of the config contained these invalid entries: %s" % ", ".join(invalid_config_items)
            raise ConfigurationError(problem)

        self.normalizeUserStory()

    # in case config is inconsistent in naming stories: e.g. in entities it is "Story", in selectors it is "HR":
    def normalizeUserStory(self):
        story_aliases = ['HierarchicalRequirement', 'UserStory', 'Story']
        story = [entity for entity in self.entities if entity in story_aliases]
        if story and self.selectors:
            story_selector = [{k:v} for s in self.selectors for k,v in s.items() if k in story_aliases]
            name = [k for s in self.selectors for k in s.keys()]
            if story_selector and story[0] not in story_selector[0]:
                name  = name[0]
                value = {}
                value[story[0]] = story_selector[0][name]
                temp = [x for x in self.config['Selectors'] for k in x.keys() if k != name]
                temp.append(value)
                self.config['Selectors'] = temp
                self.selectors = temp

    def connect(self):
        if self.proxy:
            self.log.info("Proxy for AgileCentral connection:  %s" % self.proxy)
            os.environ['HTTPS_PROXY'] = self.proxy

        self.log.info("Connecting to AgileCentral")
        custom_headers = self.get_custom_headers()

        try:
            self.ac = Rally(self.server, user=self.username, password=self.password, apikey=self.apikey,
                            version=self.rallyWSAPIVersion(), headers=custom_headers,
                            server_ping=False,
                            logger=self.restapi_logger, warn=False, debug=True)
        except Exception as msg:
            raise ConfigurationError("Unable to connect to Agile Central at %s: %s" % (self.server, msg))
        self.log.info("Connected to Agile Central server: %s" % self.server)
        all_workspaces = self.ac.getWorkspaces()
        valid = [wksp for wksp in all_workspaces if wksp.Name == self.workspace_name]
        if not valid:
            problem = "Specified Workspace: '%s' not in list of workspaces " + \
                      "available for your credentials"
            #raise ConfigurationError(problem % (self.workspace_name, self.username))
            raise ConfigurationError(problem % self.workspace_name)
        wksp = self.ac.getWorkspace()
        if wksp.Name != self.workspace_name:
            self.ac.setWorkspace(self.workspace_name)
        wksp = self.ac.getWorkspace()
        self.log.info("    Workspace: %s" % self.workspace_name)
        prjt = self.ac.getProject()
        if self.project_name:
            self.log.info("    Project  : %s" % prjt.Name)

        self.workspace_ref = wksp.ref
        self.project_ref = prjt.ref

        # detect any duplicate project names
        # self.project_bucket = {}
        # for proj in response:
        #     if proj.Name not in self.project_bucket:
        #         self.project_bucket[proj.Name] = 0
        #     self.project_bucket[proj.Name] += 1
        # self.duplicated_project_names = [p for p, c in self.project_bucket.items() if c != 1]

        #project_names = [proj.Name for proj in response]
        #self.log.info("    %d sub-projects" % len(project_names))

        return True

    def validateProjects(self, workspace, config_projects):
        """
            make requests to AgileCentral to retrieve basic info for each project in target_projects.
            If any project name in target_projects does NOT have a corresponding project in AgileCentral
            raise and Exception naming the offending project.
        """
        target_projects = [project.replace('/*', '') for project in config_projects]
        try:
            mep_projects     = list(set([project for project in target_projects if ' // '     in project]))
        except Exception as msg:
            self.log.error("%s %s" % ("error in agile_central_connection.py for getting multi-element path projects", msg))
            return False
        try:
            non_mep_projects = list(set([project for project in target_projects if ' // ' not in project]))
        except Exception as msg:
            self.log.error("%s %s" % ("error in agile_central_connection.py for getting non multi-element path projects", msg))
            return False
        query = self._construct_ored_Name_query(non_mep_projects)
        response = self.ac.get('Project', fetch='Name,ObjectID', query=query, workspace=workspace,
                                   project=None, projectScopeDown=True, pagesize=200)
        if response.errors or response.resultCount == 0:
            self.log.error( 'Unable to obtain Project name(s) specified in the config for the target Workspace: %s' % (workspace))
            return False

        found_projects = [project for project in response]
        try:
            found_project_names = list(set([p.Name for p in found_projects]))
        except Exception as msg:
            self.log.error("%s %s" % ("error in agile_central_connection.py for getting found_project_names", msg))
            return False
        bogus = [name for name in target_projects if name not in found_project_names]
        try:
            real_bogus = set(bogus) - set(mep_projects)
        except Exception as msg:
            self.log.error("%s %s" % ("error in agile_central_connection.py for getting real_bogus", msg))
            return False
        if real_bogus:
            problem =  "The project(s) mentioned in the config were not located in AgileCentral Workspace %s: %s. " % (workspace, ",".join(real_bogus))
            problem += "The project(s) may be closed, do not exist or the user may not have sufficent permissions"
            self.log.error(problem)
            return False
        if len(found_project_names) != len(found_projects) and len(mep_projects) == 0:
            duped_project_names = [name for name in found_project_names if len([project for project in found_projects if project.Name == name]) > 1]
            problem = "These Project names are ambiguous (%s).  There are multiple instances of these project names in your Workspace.  " % ", ".join(duped_project_names)
            suggestion = "The Project names must be disambiguated using this syntax: TopProject // SubProject // YourTargetProject"
            self.log.error(problem + suggestion)
            return False


        for mep in mep_projects:
            good = True
            try:
                if not self.ac.getProject(mep):
                    good = False
            except:
                good = False
            if not good:
                problem = 'Unable to obtain Project %s specified in the config for the target Workspace: %s' % (mep, workspace)
                self.log.error(problem)
                return False

        return True


    def validate(self):
        try:
            for chunk in self.config['Workspaces']:
                if 'Projects' in chunk:
                    if not self.validateProjects(chunk['Workspace'], chunk['Projects']):
                        self.log.fatal("ConfigurationError detected, see specific description above.")
                        return False

            self.entities, invalid_entities = self.vetConfiguredEntities()
            if not self.entities:
                self.log.error("Entities have no valid entries")
                return False
            if invalid_entities:
                self.log.error("Entities have bad entries: %s" % ",".join(invalid_entities))
                return False
            if 'ObjectID' not in self.fetch:
                self.fetch = 'ObjectID,' + self.fetch
            if not self.fetch.startswith('ObjectID'):
                fetch_items = self.fetch.split(',')
                if 'ObjectID' in fetch_items: fetch_items.remove('ObjectID')
                self.fetch = 'ObjectID,' + ",".join(fetch_items)
            if 'WorkspacePermission' in self.entities or 'ProjectPermission' in self.entities or 'UserPermission' in self.entities:
                if 'CustomObjectID' not in self.fetch:
                    self.fetch = 'CustomObjectID,' + self.fetch
                if not self.fetch.startswith('CustomObjectID'):
                    fetch_items = self.fetch.split(',')
                    if 'ObjectID' in fetch_items: fetch_items.remove('CustomObjectID')
                    self.fetch = 'CustomObjectID,' + ",".join(fetch_items)

            temp_fetch = []
            for item in self.fetch.split(","):
                if item not in [element for element in temp_fetch]:
                    temp_fetch.append(item)
            self.fetch = ",".join(temp_fetch)
            if not self.validateBaseQuery():
                return False
            if not self.validateSelectors():
                return False
            return True
        except Exception as e:
            self.log.write(Tracer(e).getTrace())
            return False


    def vetConfiguredEntities(self):
        goods, bads = [], []
        for entity in self.entities:
            # an entity might have a space in it, but in order to get the typedef
            # we have to use a value with no spaces (ie/, any existing spaces squished out)
            # but we need to do that locally and not mutate any of the items in self.entities
            # as we want to avoid unintended side effects to other utilizers of items in self.entities
            result = self.ac.typedef(entity[:].replace(" ",""))
            if not result:
                bads.append(entity)
            else:
                goods.append(entity.replace('PortfolioItem/',''))
        return goods, bads

    def findQueryAttributes(self, query_conditions):

        RELATIONAL_OPS = "|".join(['=', '>', '<', '<=', '>=', "contains"])
        CONJUNCTORS = '(\) ?AND ?\(|\) ?OR ?\()'
        conj_re = re.compile(CONJUNCTORS, flags=re.I)
        base_attributes_found = []

        for condition in [cond for cond in query_conditions if cond]:
            target = condition[:]
            if target[0] == "(" and target[-1] == ")" and target[:2] == "((" and target[-2:] == "))":
                target = target[1:-1]

            target = re.sub(r'!', '', target)  # scrub any '!' character in the string
            target = re.sub(r'\(+', '(', target)  # turn multiple ((( chars into a single '('
            target = re.sub(r'\)+', ')', target)  # turn multiple ))) chars into a single ')'

            expressions = [item.lstrip('(') for item in conj_re.split(target) if not conj_re.search(item.strip())]

            for expr in expressions:
                attr_full = re.split(r'(%s)' % RELATIONAL_OPS, expr)[0].strip()
                base_attr = attr_full.split('.')[0]
                if base_attr not in base_attributes_found:
                    base_attributes_found.append(base_attr)

        return base_attributes_found


    def validateBaseQuery(self):
        for entity in self.entities:
            base_query_field_names = self.findQueryAttributes([self.base_query])
            if base_query_field_names:
                no_spaces_entity = entity[:].replace(" ", "")
                valid_fields = [field.ElementName for field in self.ac.typedef(no_spaces_entity).Attributes]
                invalid_fields = [field for field in base_query_field_names if field not in valid_fields]
                if invalid_fields:
                    problem = "The following attributes used in your BaseQuery for %s do not exist: '%s'" % (
                        entity, ",".join(invalid_fields))
                    self.log.error(problem)
                    return False
        return True

    def validateSelectors(self):
        if not self.selectors:
            return True
        # in case config is inconsistent in naming stories: e.g. in entities it is "Story", in selectors it is "HR":
        possible_invalid_entities = [entity for selector in self.selectors for entity in selector.keys() if entity not in self.entities]
        story_aliases = ['HierarchicalRequirement', 'UserStory', 'Story']
        invalid_entities = [entity for entity in possible_invalid_entities if entity not in story_aliases and entity not in self.entities]
        #invalid_entities = [entity for selector in self.selectors for entity in selector.keys() if entity not in self.entities]
        if invalid_entities:
            problem = "Detected invalid Entities: %s" % ", ".join(invalid_entities)
            self.log.error(problem)
            return False

        problems = []
        valid_fields = []
        for selector in self.selectors:
            for entity in selector.keys():
                if selector[entity].__class__.__name__ == 'str':
                    conditions = [selector[entity]]
                elif selector[entity].__class__.__name__ == 'list':
                    conditions = selector[entity]
                else:
                    problem = "Unrecognized value in Selector for %s" % entity
                    raise ConfigurationError(problem)
                target_attrs = self.findQueryAttributes(conditions)
                if target_attrs:
                    valid_fields = [field.ElementName for field in self.ac.typedef(entity).Attributes]
                invalid_fields = [field for field in target_attrs if field not in valid_fields]
                if invalid_fields:
                    problem = "The following attributes used in your selector condition for %s do not exist: '%s'" % (
                            entity, ",".join(invalid_fields))
                    problems.append(problem)
        if problems:
            self.log.error(problems)
            return False
        return True


    def constructWsapiQuery(self, base, esq):
        #esq stands for entity selector query (can either be a single string or a list of strings)
        result = None
        if base:
            if base[0] != '(' and base[-1] != ')':
                base = "(%s)" % base
        if not esq:
            return base

        def addAnAndedCondition(current, addition):
            return addition if not current else "(%s AND %s)" % (current, addition)

        if esq.__class__.__name__ == 'str':
            if esq[0] != '(' and esq[-1] != ')':
                esq = "(%s)" % esq
            result = addAnAndedCondition(base, esq)
        elif esq.__class__.__name__ == 'list' and esq:
            esq = [e if e[0] == '(' and e[-1] == ')' else "(%s)" % e for e in esq]
            result = addAnAndedCondition(base, esq[0])
            for e in esq[1:]:
                result = addAnAndedCondition(result, e)
        return result


    def assembleFullQuery(self, entity):
        conditions = []
        ted = [d for d in self.selectors if entity in d] if self.selectors else None
        if ted:
            condition = ted[0][entity]
            if condition.__class__.__name__ == 'str':
                conditions.append(condition)
            elif condition.__class__.__name__ == 'list':
                conditions.extend(condition)

        result = self.constructWsapiQuery(self.base_query, conditions)
        return result


    def getSchema(self, entity):
        full_schema = self.ac.typedef(entity[:].replace(" ",""))
        valid_attributes = list(filter(self.getSpecifiedAttributes, full_schema.Attributes))
        partial_schema = []
        eligible_fields = self.fetch.split(',')
        if entity in ['WorkspacePermission', 'ProjectPermission', 'UserPermission']:
            if 'ObjectID' in eligible_fields: eligible_fields.remove('ObjectID')  # ObjectID is null in permission entities
        for elementName in eligible_fields:
            hits = [attr for attr in valid_attributes if attr.ElementName == elementName]
            if hits:
                partial_schema.append(hits[0])

        if self.config.get('ResolveUser', False):
            [setattr(attr, 'AttributeType', 'STRING') for attr in partial_schema if attr.SchemaType == 'User']

        def sad(sa):
            """
                sad --> SchemaAttributeDisplay  sa --> SchemaAttribute item
                   return a string of the following form: 'ElementName: %s   Type: %s   Custom: Yes/No   Constrained: Yes/No'
                   followed by a string or strings of the AllowedValues if there are any
            """
            custom      = 'Yes' if sa.Custom else 'No'
            constrained = 'Yes' if sa.Constrained else 'No'
            ident = "ElementName: %s   Type: %s   Custom: %s   Constrained: %s" % (sa.ElementName, sa.AttributeType, custom, constrained)

            if hasattr(sa, 'AllowedValueType') and sa.AllowedValueType and not sa._allowed_values_resolved:
                avt_ref = "/".join(sa.AllowedValueType['_ref'].split('/')[-2:])
                av_line = "AllowedValues:  unresolved_ref: %s" % avt_ref
                return "\n".join(["       " + ident, "                 " + av_line])
            if hasattr(sa, 'AllowedValueType') and hasattr(sa, 'AllowedValues') and sa.AllowedValues and type(sa.AllowedValues) == list:
                avs = []
                for ix, item in enumerate(sa.AllowedValues):
                    if type(item) == str:
                        avs.append(item)
                    elif type(item) == dict:
                        avs.append(str(item['StringValue']))
                    else:
                        avs.append(str(item.__dict__['StringValue']))
                avv_line        = "                AllowedValues: %s" % repr(avs)
                return "\n".join(["        " + ident, avv_line])

            # default to just returning the ident line for this Attribute
            return "        " + ident


        state_attrs  = [attr for attr in partial_schema if attr.ElementName == 'State']
        state_attrs_details = "\n    -----------\n    ".join([sad(sa) for sa in state_attrs])
        self.log.debug("ACConn.getSchema(%s) attrs whose ElementName is 'State':\n    %s" % (entity, state_attrs_details))
        object_attrs = [attr for attr in partial_schema if attr.AttributeType == 'OBJECT']
        object_attrs_details = "\n    ".join([sad(oa) for oa in object_attrs])
        self.log.debug("ACConn.getSchema(%s) attrs whose AttributeType is OBJECT:\n    %s" % (entity, object_attrs_details))

        pi_state_attrs = [attr for attr in partial_schema if
                          attr.AttributeType == 'OBJECT' and attr.ElementName == 'State']

        self.log.debug("ACConn.getSchema(%s) pi_state_attrs: %s" % (entity, ", ".join([sa.ElementName for sa in pi_state_attrs])))
        if pi_state_attrs:
            self.log.debug("Inside pi_state_attrs...")
            self.ac.enableLogging(dest=b'pyral.debug.log', attrget=True)
            pi_states = [item for item in self.ac.getStates(entity)]
            pi_state_values = [state.Name for state in pi_states]
            self.log.debug("  pi_state_values: %s" % ", ".join(pi_state_values))
            pi_state_attr = pi_state_attrs[0]
            pi_state_attr.AllowedValues = pi_state_values
            self.pi_states_map[entity] = {state.oid: state.Name for state in pi_states}
            self.log.debug("ACConn.getSchema(%s) .pi_states_map[%s]: %s" % (entity, entity, self.pi_states_map[entity]))
            self.ac.disableLogging()
        partial_schema_idents = ", ".join(['SchemaItemAttribute for %s' % si.ElementName for si in partial_schema])
        self.log.debug("ACConn.getSchema(%s) returning partial_schema of: [%s]" % (entity, partial_schema_idents))
        return partial_schema


    def _construct_ored_Name_query(self, target_items):
        if not target_items: return ''
        initial = '(Name = "%s")' % target_items[0]
        or_string = initial
        for item in target_items[1:]:
            or_string = '(%s OR (Name = "%s"))' % (or_string, item)
        return or_string


    def getSpecifiedAttributes(self, element):
        found = False
        result = re.search('\\b%s\\b' % element.ElementName, self.fetch)
        if result:
            found = True
        return found


    def entityAttributes(self, entity, entity_schema):
        self.log.debug("Getting schema for AC entity: %s" % entity)
        attribute_names = ", ".join([attr.ElementName for attr in entity_schema])
        self.log.debug("Obtained schema for AC entity: %s, narrowed to: %s " % (entity, attribute_names))
        columns = [{attr.ElementName: (attr.AttributeType, attr.SchemaType)} for attr in entity_schema]
        return columns

    # if you use this to resolve User to UserName, be aware that it greatly increases processing time as a call will be made to AC for each user valued field
    # def get_username(self, oid):
    #     fields = 'ObjectID,UserName'
    #     query = 'ObjectID = %s' % oid
    #     response = self.ac.get('User', fetch=fields, query=query, pagesize=200)
    #     if response.resultCount:
    #         return [user for user in response][0]

    def resolveFieldValue(self, entity, field, raw_value, category, remapping):
        # RATING   type e.g. Severity     when empty return 'None'.
        # QUANTITY type e.g. PlanEstimate when empty return None
        value = raw_value  # be optomistic that no transform on the raw_value is necessary for this field
        if raw_value in [None, '', 'None', []]:
            value = None
        else:
            if field in category['user']:
                if self.config['ResolveUser']:
                    value = raw_value.Name
                else:
                    value = int(raw_value._ref.split('/')[-1])
            elif field in category['ref']:
                value = int(raw_value._ref.split('/')[-1])
            elif field in category['pi_state']:
                oid = raw_value._ref.split('/')[-1]
                value = self.pi_states_map[entity].get(int(oid))
            elif raw_value.__class__.__name__ == 'list':
                if field == 'UserPermissions':
                    value = set([item.CustomObjectID for item in raw_value])
                else: # any other collection except 'UserPermission'
                    value = set([item.ObjectID for item in raw_value])
            elif remapping:
                value = remapping.get(value, value)
            elif field == 'FormattedID' or (field not in category['number'] and field != 'None'):
                value = "'" + str(raw_value) + "'"

        return value


    def createFieldCategories(self, entity, entity_schema):
        columns = self.entityAttributes(entity, entity_schema)

        fields            = [k for column in columns for k, v in column.items()]
        ref_fields        = [k for column in columns for k, v in column.items()
                                if (v[0] == 'OBJECT' and k != 'State' and v[1] != 'User')]
        pi_state_fields   = [k for column in columns for k, v in column.items()
                                if (v[0] == 'OBJECT' and k == 'State')]
        user_fields       = [k for column in columns for k, v in column.items() if v[1] == 'User']
        collection_fields = [k for column in columns for k, v in column.items()
                               for v in column.values() if 'COLLECTION' in v]
        number_fields     = [k for column in columns for k, v in column.items()
                               for v in column.values() if 'INTEGER' in v or 'QUANTITY' in v or 'DECIMAL' in v]
        boolean_fields    = [k for column in columns for k, v in column.items()
                               for v in column.values() if 'BOOLEAN' in v]

        field_categories = \
            {'ref'          : ref_fields,
             'pi_state'     : pi_state_fields,
             'user'         : user_fields,
             'number'       : number_fields,
             'collection'   : collection_fields,
             'boolean'      : boolean_fields
             }
        self.log.debug("ACConn.createFieldCategories  has pi_state_fields: %s  and field_categories['pi_state'] of: %s" % \
                       (repr(pi_state_fields), repr(field_categories['pi_state'])))
        return fields, field_categories

    def getQualifierCount(self, entity, workspace, project, scope_up, scope_down):
        entity_schema, fields, field_categories, response = self.getInitialQualifier(workspace, project, entity, scope_up, scope_down, 1)
        return response.resultCount


    def getInitialQualifier(self, workspace, project, entity, scope_up, scope_down, pagesize):
        self.ac.setWorkspace(workspace)
        if project:
            self.ac.setProject(project)
        entity_schema = self.getSchema(entity)

        try:
            self.log.debug('Retrieving Agile Central items for %s' % entity)
            fields, field_categories = self.createFieldCategories(entity, entity_schema)
            #if entity in ['WorkspacePermission','ProjectPermission', 'UserPermission']:
            #    fields.remove('ObjectID') # ObjectID is null in permission entities
            fetch = ','.join(fields)
            full_query = self.assembleFullQuery(entity)
            response = self.ac.get('%s' % entity[:].replace(" ",""), fetch=fetch, query=full_query, order="ObjectID", project=project, \
                                   pagesize=pagesize, projectScopeUp=scope_up, projectScopeDown=scope_down)
            if response.errors:
                raise OperationalError(response.errors[0])
        except Exception as e:
            problem = "Unable to retrieve AgileCentral data for entity: %s, exception type: %s" % (
            entity, type(e).__name__)
            raise OperationalError(problem)
        return entity_schema, fields, field_categories, response


    def indexBy(self, entity):
        if entity in ['WorkspacePermission','ProjectPermission', 'UserPermission']:
            return{
                'WorkspacePermission': 'customobjectid',
                'ProjectPermission'  : 'customobjectid',
                'UserPermission'     : 'customobjectid'
            }[entity]
        else:
            return 'ObjectID'


    def getAgileCentralData(self, workspace, project, entity, scope_up, scope_down, remappings):
        self.cached_field_type = {}
        records = []
        pagesize = 1000
        def getFieldType(field_name, entity_schema):
            if field_name in self.cached_field_type:
                return self.cached_field_type[field_name]
            field_type = [si.AttributeType for si in entity_schema if si.ElementName == field_name][0].lower()
            self.cached_field_type[field_name] = field_type
            return field_type

        entity_schema, fields, field_categories, response = self.getInitialQualifier(workspace, project, entity, scope_up, scope_down, pagesize)
        for item in response:
            field_values = []
            try:
                for field in fields:
                    field_type = getFieldType(field, entity_schema)
                    raw_value = getattr(item, field)
                    value = self.resolveFieldValue(entity, field, raw_value, field_categories, remappings.get(field_type, {}))
                    field_values.append(value) # None values are also appended due to indentation
            except:
                e = sys.exc_info()[0]
                self.log.error("Problem in fields loop for %s with OID of %s  field name: %s, skipping item\n%s" % (entity, item.oid, field, e))
                continue
            records.append(tuple(field_values)) # example: [("'E863'", 4),{"'E865'",6)] where 4 is DirectChildrenCount
        self.log.debug("Retrieved %d records for %s" % (len(records), entity))

        statistics = {'Number of Records': response.resultCount, 'Fields Per Record': len(fields)}

        pk_field = self.indexBy(entity)
        return entity_schema, pk_field, fields, records, statistics

